#include "string_encoding.hpp"

#include <QDebug>

#ifdef QTTEXMACS
#include <QString>
//#define USE_QT_STRING_CONVERSION
#endif

#ifndef USE_QT_STRING_CONVERSION
#include <windows.h>
#endif

UINT get_code_page() {
  UINT code_page = GetACP();
  return code_page;
}

// the input string in in utf8. the output string will be used for fopen
// according to the windows documentation : 
// " By default, a narrow filename string is interpreted using the ANSI codepage (CP_ACP). In Windows Desktop applications, it can be changed to the OEM codepage (CP_OEMCP) by using the SetFileApisToOEM function."
// source : https://learn.microsoft.com/en-us/cpp/c-runtime-library/reference/fopen-wfopen?view=msvc-170
// The ANSI code pages can be different on different computers, or can be changed for a single computer, leading to data corruption.
// source : https://learn.microsoft.com/en-us/windows/win32/api/winnls/nf-winnls-getacp
std::string texmacs_utf8_string_to_system_string(string utf8_string) {
  // step 1 : convert the string to multibyte wide char
  c_string _utf8_string = utf8_string;
  int wide_size_needed = MultiByteToWideChar(CP_UTF8, 0, _utf8_string, N(utf8_string), NULL, 0);
  if (wide_size_needed == 0) {
    return "";
  }
  std::wstring wide_string(wide_size_needed, 0);
  MultiByteToWideChar(CP_UTF8, 0, _utf8_string, N(utf8_string), &wide_string[0], wide_string.size());

  // step 2 : convert the wide char to ainsi
  int ansi_size_needed = WideCharToMultiByte(get_code_page(), 0, wide_string.c_str(), wide_string.size(), NULL, 0, NULL, NULL);
  if (ansi_size_needed == 0) {
    return "";
  }
  std::string ansi_string(ansi_size_needed, 0);
  WideCharToMultiByte(get_code_page(), 0, wide_string.c_str(), wide_string.size(), &ansi_string[0], ansi_string.size(), NULL, NULL);
  return ansi_string;
}



std::wstring texmacs_utf8_string_to_system_widestring(string utf8_string) {
  c_string _utf8_string = utf8_string;
  int wide_size_needed = MultiByteToWideChar(CP_UTF8, 0, _utf8_string, N(utf8_string), NULL, 0);
  if (wide_size_needed == 0) {
    return L"";
  }
  std::wstring wide_string(wide_size_needed, 0);
  MultiByteToWideChar(CP_UTF8, 0, _utf8_string, N(utf8_string), &wide_string[0], wide_string.size());
  return wide_string;
}



// local8bit_string is in ainsi
string texmacs_system_string_to_utf8_string(const std::string &local8bit_string) {
  // step 1 : convert the string to multibyte wide char
  int wide_size_needed = MultiByteToWideChar(get_code_page(), 0, local8bit_string.c_str(), local8bit_string.size(), NULL, 0);
  if (wide_size_needed == 0) {
    return "";
  }
  std::wstring wide_string(wide_size_needed, 0);
  MultiByteToWideChar(get_code_page(), 0, local8bit_string.c_str(), local8bit_string.size(), &wide_string[0], wide_string.size());

  // step 2 : convert the wide char to utf8
  int utf8_size_needed = WideCharToMultiByte(CP_UTF8, 0, wide_string.c_str(), wide_string.size(), NULL, 0, NULL, NULL);
  if (utf8_size_needed == 0) {
    return "";
  }
  std::string utf8_string(utf8_size_needed, 0);
  WideCharToMultiByte(CP_UTF8, 0, wide_string.c_str(), wide_string.size(), &utf8_string[0], utf8_string.size(), NULL, NULL);
  return {utf8_string.data(), utf8_string.size()};
}



string texmacs_system_wstring_to_utf8_string(const std::wstring &local8bit_string) {
  int utf8_size_needed = WideCharToMultiByte(CP_UTF8, 0, local8bit_string.c_str(), local8bit_string.size(), NULL, 0, NULL, NULL);
  if (utf8_size_needed == 0) {
    return "";
  }
  std::string utf8_string(utf8_size_needed, 0);
  WideCharToMultiByte(CP_UTF8, 0, local8bit_string.c_str(), local8bit_string.size(), &utf8_string[0], utf8_string.size(), NULL, NULL);
  return {utf8_string.data(), utf8_string.size()};
}

char *c_texmacs_utf8_string_to_system_string(const char *utf8_string) {
  std::string system_string = texmacs_utf8_string_to_system_string(utf8_string);
  char *c_system_string = (char*)malloc(system_string.size() + 1);
  memcpy(c_system_string, system_string.c_str(), system_string.size());
  c_system_string[system_string.size()] = 0;
  return c_system_string; // warning : the memory shoud be deleted
}



char *c_system_string_to_utf8_string(const char *system_string) {
  string utf8_string = texmacs_system_string_to_utf8_string(system_string);
  char *c_utf8_string = (char*)malloc(N(utf8_string) + 1);
  memcpy(c_utf8_string, &utf8_string[0], N(utf8_string));
  c_utf8_string[N(utf8_string)] = 0;
  return c_utf8_string; // warning : the memory shoud be deleted
}



void texmacs_utf8_string_to_system_string_path(char *utf8_string) {
  std::string system_string = texmacs_utf8_string_to_system_string(utf8_string);
  if (system_string.size() >= MAX_PATH - 1) {
    // do nothing
    return;
  }
  memcpy((void*)utf8_string, system_string.c_str(), system_string.size());
  utf8_string[system_string.size()] = 0;
}