#ifndef TEXMACS_WINDOWS_STRING_ENCODING_HPP
#define TEXMACS_WINDOWS_STRING_ENCODING_HPP

#include "string.hpp"
#include <string>

/**
  * @brief Convert a UTF-8 string to a string with the system encoding.
  * @param utf8_string The UTF-8 string to convert.
  * @return A string with the system encoding.
  * @see texmacs_utf8_string_to_system_string(string utf8_string)
  *
  * This function converts a UTF-8 string to a string with the locale system encoding.
  * The returned string can be used with the standard C functions,
  * but should not be used for texmacs document contents.
  */
std::string texmacs_utf8_string_to_system_string(string utf8_string);

/**
  * @brief The wide version of @ref texmacs_utf8_string_to_system_string.
  * @param utf8_string The UTF-8 string to convert.
  * @return A wide string with the system encoding.
  * @see texmacs_utf8_string_to_system_string(string utf8_string)
  *
  * This function converts a UTF-8 string to a wide string with the locale system encoding.
  * The returned string can be used with the standard C functions,
  * but should not be used for texmacs document contents.
 */
std::wstring texmacs_utf8_string_to_system_widestring(string utf8_string);

/**
  * @brief Convert a string with the system encoding to a UTF-8 string.
  * @param local8bit_string The string with the system encoding to convert.
  * @return A UTF-8 string.
  * @see texmacs_system_string_to_utf8_string(c_string local8bit_string)
  *
  * This function converts a string with the locale system encoding to a UTF-8 string.
  * The returned string should not be used with the standard C functions,
  * but can be used for texmacs document contents.
 */
string texmacs_system_string_to_utf8_string(const std::string &local8bit_string);

/**
  * @brief The wide version of @ref texmacs_system_string_to_utf8_string.
  * @param local8bit_string The string with the system encoding to convert.
  * @return A wide UTF-8 string.
  * @see texmacs_system_string_to_utf8_string(c_string local8bit_string)
  *
  * This function converts a string with the locale system encoding to a wide UTF-8 string.
  * The returned string should not be used with the standard C functions,
  * but can be used for texmacs document contents.
 */
string texmacs_system_wstring_to_utf8_string(const std::wstring &local8bit_string);

/**
  * @brief Function dedicated to Guile. C Version of @ref texmacs_utf8_string_to_system_string.
 */
char *c_texmacs_utf8_string_to_system_string(const char *utf8_string);


/**
  * @brief Function dedicated to Guile. C Version of @ref texmacs_utf8_string_to_system_widestring.
 */
char *c_system_string_to_utf8_string(const char *system_string);


/**
  * @brief Function dedicated to Guile. This is an inline C version of @ref texmacs_system_string_to_utf8_string.
  *
  * This function is necessary for char with MAX_PATH length, such as dirent.
 */
void texmacs_utf8_string_to_system_string_path(char *utf8_string);

#endif // TEXMACS_WINDOWS_STRING_ENCODING_HPP