#define SECURITY_WIN32

#include "windows_system.hpp"
#include <windows.h>
#include <shellapi.h>
#include <sspi.h>
#include <secext.h>
#include <stdio.h>

#include <QApplication>
#include <QDebug>
#include <iostream>

#include "Scheme/Guile/guile_tm.hpp"

/******************************************************************************************************
* Entry point
******************************************************************************************************/

#if defined(QT_NEEDS_QMAIN)
#error "Qt is replacing main with qMain. Please, remove the usage of QMain (you shouldn't have a QT_NEEDS_QMAIN). With --with-qt-find-method=pkg-config it is the default behavior."
#endif

int texmacs_main(int argc, char** argv);

int WINAPI CommonMain() {
    texmacs_attach_console();
    texmacs_initialize_displayname();
    texmacs_init_guile_hooks();

    // Now, we are going to convert the arguments to texmacs string UTF-8
    string *texmacs_argv = new string[__argc];
    char **char_argv = new char*[__argc];
    for (int i = 0; i < __argc; i++) {
#ifdef _UNICODE
        texmacs_argv[i] = texmacs_system_wstring_to_utf8_string(__wargv[i]);
#else
        texmacs_argv[i] = texmacs_system_string_to_utf8_string(__argv[i]);
#endif
        char_argv[i] = as_charp(texmacs_argv[i]);
    }

    int result = texmacs_main(__argc, char_argv);
    return result;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
    return CommonMain();
}

int WINAPI wWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow) {
    return CommonMain();
}

int main(int argc, char** argv) {
    return CommonMain();
}

/******************************************************************************************************
* System functions
******************************************************************************************************/
typedef struct texmacs_dir_t {
    HANDLE handle;
    WIN32_FIND_DATAW find_data;
    bool is_find_data_valid;
} texmacs_dir_t;

FILE* texmacs_fopen(string filename, string mode, bool lock) {
    std::wstring wide_filename = texmacs_utf8_string_to_system_widestring(filename);
    std::wstring wide_mode = texmacs_utf8_string_to_system_widestring(mode);
    wide_mode += L"b";
    FILE* result = _wfopen(wide_filename.c_str(), wide_mode.c_str());
    return result;
}

void texmacs_fclose(FILE *&file, bool unlock) {
    fclose(file);
    file = nullptr;
}

TEXMACS_DIR texmacs_opendir(string dirname) {
    dirname = dirname * "\\*";
    texmacs_dir_t *dir = new texmacs_dir_t;
    dir->handle = FindFirstFileW(texmacs_utf8_string_to_system_widestring(dirname).c_str(), &dir->find_data);
    if (dir->handle == INVALID_HANDLE_VALUE) {
        delete dir;
        return nullptr;
    }
    dir->is_find_data_valid = true;
    return dir;
}

void texmacs_closedir(TEXMACS_DIR &dir) {
    FindClose(dir->handle);
    delete dir;
    dir = nullptr;
}

texmacs_dirent texmacs_readdir(TEXMACS_DIR dirp) {
    texmacs_dirent dirent;
    dirent.is_valid = dirp->is_find_data_valid;
    dirent.d_name = texmacs_system_wstring_to_utf8_string(dirp->find_data.cFileName);
    dirp->is_find_data_valid = FindNextFileW(dirp->handle, &dirp->find_data);
    return dirent;
}

bool texmacs_stat(string filename, struct_stat* buf) {
    return _wstat64(texmacs_utf8_string_to_system_widestring(filename).c_str(), buf);
}

string texmacs_getenv(string variable_name) {
    std::wstring wide_variable_name = texmacs_utf8_string_to_system_widestring(variable_name);
    wchar_t *value = _wgetenv(wide_variable_name.c_str());
    if (value == nullptr) {
        return "";
    }
    return texmacs_system_wstring_to_utf8_string(value);    
}

bool texmacs_setenv(string variable_name, string new_value) {
    std::wstring wide_variable_name = texmacs_utf8_string_to_system_widestring(variable_name);
    std::wstring wide_new_value = texmacs_utf8_string_to_system_widestring(new_value);
    return _wputenv_s(wide_variable_name.c_str(), wide_new_value.c_str()) == 0;
}

// putenv
bool texmacs_putenv(string variable) {
    std::wstring wide_variable = texmacs_utf8_string_to_system_widestring(variable);
    return _wputenv(wide_variable.c_str()) == 0;
}

bool texmacs_mkdir(string dirname, int mode) {
    return CreateDirectoryW(texmacs_utf8_string_to_system_widestring(dirname).c_str(), nullptr);
}

bool texmacs_rmdir(string dirname) {
    return RemoveDirectoryW(texmacs_utf8_string_to_system_widestring(dirname).c_str());
}

bool texmacs_rename(string oldname, string newname) {
    return MoveFileW(texmacs_utf8_string_to_system_widestring(oldname).c_str(), texmacs_utf8_string_to_system_widestring(newname).c_str());
}

bool texmacs_chmod(string filename, int mode) {
    // do nothing
    return true;
}

bool texmacs_remove(string filename) {
    return DeleteFileW(texmacs_utf8_string_to_system_widestring(filename).c_str()) == 0;
}

void texmacs_attach_console() {
    // detect gdb

    // get MSYSTEM variable
    char *msystem = getenv("MSYSTEM");
    
    // if MSYSTEM is not set, we are running in a msys environment
    if (msystem != nullptr && !IsDebuggerPresent()) {
        // in the msys environment, we don't need to attach to the console
        return;
    }

    // attach to an existing windows command prompt (if any)
    if (AttachConsole(ATTACH_PARENT_PROCESS)) {
        // redirect the standard input, output and error to the console
        freopen("CONIN$", "r", stdin);
        freopen("CONOUT$", "w", stdout);
        freopen("CONOUT$", "w", stderr);
    }
}

#ifdef SCM_HAVE_HOOKS
int texmacs_guile_stat(const char *path, guile_stat_t *buf) {
    std::wstring wide_path = texmacs_utf8_string_to_system_widestring(path);
    int result = _wstat64(wide_path.c_str(), buf);
    return result;
}

int texmacs_guile_lstat(const char *path, guile_stat_t *buf) {
    std::wstring wide_path = texmacs_utf8_string_to_system_widestring(path);
    return _wstat64(wide_path.c_str(), buf);
}

int texmacs_guile_open(const char *pathname, int flags, mode_t mode) {
    std::wstring wide_path = texmacs_utf8_string_to_system_widestring(pathname);
    int result = _wopen(wide_path.c_str(), flags, mode);
    return result;
}

DIR *texmacs_guile_opendir(const char *name) {
    return (DIR*)texmacs_opendir(name);
}

guile_dirent_t *texmacs_guile_readdir(DIR *_dirp) {
    texmacs_dir_t *dirp = (texmacs_dir_t*)_dirp;
    if (dirp->is_find_data_valid == false) {
        return nullptr;
    }
    guile_dirent_t *dirent = (guile_dirent_t*)malloc(sizeof(guile_dirent_t));
    string name = texmacs_system_wstring_to_utf8_string(dirp->find_data.cFileName);
    c_string c_name = name;
    strncpy(dirent->d_name, c_name, 256);
    dirp->is_find_data_valid = FindNextFileW(dirp->handle, &dirp->find_data);
    return dirent;
}

int texmacs_guile_truncate(const char *path, guile_off_t length) {
    std::wstring wide_path = texmacs_utf8_string_to_system_widestring(path);
    HANDLE file = CreateFileW(wide_path.c_str(), GENERIC_WRITE, 0, nullptr, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, nullptr);
    if (file == INVALID_HANDLE_VALUE) {
        return -1;
    }
    LARGE_INTEGER li;
    li.QuadPart = length;
    if (!SetFilePointerEx(file, li, nullptr, FILE_BEGIN)) {
        CloseHandle(file);
        return -1;
    }
    if (!SetEndOfFile(file)) {
        CloseHandle(file);
        return -1;
    }
    CloseHandle(file);
    return 0;
}

char *texamcs_guile_getenv(const char *name) {
    string utf8_string = texmacs_getenv(name);
    char *c_utf8_string = (char*)malloc(N(utf8_string) + 1);
    memcpy(c_utf8_string, &utf8_string[0], N(utf8_string));
    c_utf8_string[N(utf8_string)] = 0;
    return c_utf8_string; // warning : the memory shoud be deleted
}
#endif

void texmacs_init_guile_hooks() {
#ifdef SCM_HAVE_HOOKS
    guile_utf8_string_to_system_string = c_texmacs_utf8_string_to_system_string;
    guile_system_string_to_utf8_string = c_system_string_to_utf8_string;
    guile_utf8_string_to_system_string_path = texmacs_utf8_string_to_system_string_path;

    guile_stat = texmacs_guile_stat;
    guile_lstat = texmacs_guile_lstat;
    guile_open = texmacs_guile_open;
    guile_opendir = texmacs_guile_opendir;
    guile_readdir = texmacs_guile_readdir;
    guile_truncate = texmacs_guile_truncate;
    guile_getenv = texamcs_guile_getenv;
#endif
}

void texmacs_initialize_displayname() {
    // if TEXMACS_DISPLAYNAME is already set, we don't need to set it
    if (getenv("TEXMACS_DISPLAYNAME") != nullptr) {
        return;
    }
    // get the full username using GetUserNameExW
    wchar_t username[256 + 1];
    DWORD username_size = 256 + 1;
    bool result = GetUserNameExW(NameDisplay, username, &username_size);
    if (result) {
        // set the environment variable
        std::wstring full_username = L"TEXMACS_DISPLAYNAME=" + std::wstring(username);
        _wputenv(full_username.c_str());
    } else {
        // format the error message
        char error_message[256];
        FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM, nullptr, GetLastError(), 0, error_message, 256, nullptr);
        cout << "texmacs_initialize_displayname error " << error_message << "\r\n";
        // set to Default User
        _putenv("TEXMACS_DISPLAYNAME=Default User");
    }
}

intptr_t texmacs_spawnvp(int mode, string name, array<string> args) {
    // convert the arguments to a wide string
    std::vector<wchar_t*> wide_args;
    for (int i = 0; i < N(args); i++) {
        std::wstring wide_arg = texmacs_utf8_string_to_system_widestring(args[i]);
        wchar_t *c_wide_arg = (wchar_t*)malloc((wide_arg.size() + 1) * sizeof(wchar_t));
        memcpy(c_wide_arg, &wide_arg[0], wide_arg.size() * sizeof(wchar_t));
        c_wide_arg[wide_arg.size()] = 0;
        wide_args.push_back(c_wide_arg);
    }

    // convert the name to a wide string
    std::wstring wide_name = texmacs_utf8_string_to_system_widestring(name);

    // spawn the process
    intptr_t res = _wspawnvp(mode, wide_name.c_str(), (wchar_t* const*)wide_args.data());

    // free the memory
    for (int i = 0; i < wide_args.size(); i++) {
        free(wide_args[i]);
    }

    return res;
}